const synaptic = require('synaptic')
const _ = require('lodash')

const Crossover = require('./crossover')
const Mutation = require('./mutation')

class Evonet {

  constructor (settings) {
    settings = settings || {}
    this.population = settings.population
    this.winners = settings.winners || Math.round(this.population * 0.5)
    this.mutationRate = settings.mutationRate
    this.layers = settings.layers

    this.reset()
  }

  /**
   * Reset this evolutionary network.
   */
  reset () {
    this.generation = 1
    this.networks = []
    this.best = { fitness: 0 }
    for (let i = 0; i < this.population; ++i) {
      this.networks[i] = {
        network: new synaptic.Architect.Perceptron(...(this.layers)),
        fitness: 0
      }
    }
  }

  /**
   * Activate a particular network and feed in sensor input.
   * @param {Number} n Which network to activate (0 <= n < population).
   * @param {Number[]} input An input of pre-defined size (as specified by the input layer size).
   */
  activate (n, input) {
    return this.networks[n].network.activate(input)
  }

  /**
   * Evaluate a particular network and provide it's fitness score.
   * @param {Number} n Which network to evaluate (0 <= n < population).
   * @param {Number} fitness A fitness score (higher is better).
   */
  evaluate (n, fitness) {
    this.networks[n].fitness = fitness
  }

  /**
   * Evolve this generation and produce a new generation of networks.
   */
  evolve () {
    const winners = this.networks.sort((a, b) => b.fitness - a.fitness).slice(0, this.winners)
    if (this.best.fitness < winners[0].fitness) this.best = winners[0]
    for (let i = this.winners; i < this.population; ++i) {
      let parents = _.sampleSize(winners, 2)
      let offspringNetwork = Crossover.standard(parents[0].network.toJSON(), parents[1].network.toJSON())
      offspringNetwork = Mutation.standard(offspringNetwork, this.mutationRate)
      this.networks[i] = {
        network: synaptic.Network.fromJSON(offspringNetwork),
        fitness: 0
      }
    }
  }

}

module.exports = Evonet
